# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module for public API on PyWorkbench."""

import atexit
import logging
import tempfile

from ansys.workbench.core.workbench_client import WorkbenchClient
from ansys.workbench.core.workbench_launcher import Launcher


class ClientWrapper(WorkbenchClient):
    """Provides for connecting to a Workbench server.

    Parameters
    ----------
    port : int
        Port used by the server.
    client_workdir : str, default: None
        Path to a writable directory on the client computer.
    host : str, default: None
        Server computer's name or IP address.
    security : str, default: 'mtls'
        Transport mode used for connection security.
        Options are: "insecure", "uds", "wnua", "mtls"
    """

    def __init__(self, port, client_workdir=None, host=None, security="mtls"):
        """Create a PyWorkbench client that connects to a Workbench server."""
        if host is None:
            host = "localhost"
        if client_workdir is None:
            client_workdir = tempfile.gettempdir()
        super().__init__(client_workdir, host, port, security)
        super()._connect()

    def exit(self):
        """Disconnect from the server."""
        if super()._is_connected():
            super()._disconnect()
            logging.info("Workbench server connection has ended.")


class LaunchWorkbench(ClientWrapper):
    """Launch a Workbench server on a local machine or a remote Windows machine.

    Parameters
    ----------
    show_gui : bool, default: True
        Weather to launch Workbench in UI mode.
    version : str, default: None
        Workbench version to launch. It must be a 3-digit version that is "242" or later.
    client_workdir : str, default: None
        Path to a writable directory on the client computer. The default is ``None``,
        in which case the system temp directory is used.
    server_workdir : str, None
        Path to a writable directory on the server computer. The default is ``None``,
        in which case the user preference for the Workbench temporary file folder is used.
    use_insecure_connection : bool, default: False
        whether to use insecure connection between the server and clients
    host : str, None
        Server computer's name or IP address. The default is ``None`` for launching on the
        local computer.
    username : str, None
        User's login name on the server computer. The default is ``None`` for launching on
        the local computer.
    password : str, None
        User's password on the server computer. The default is ``None`` for launching on
        the local computer.

    Raises
    ------
    Exception
        If PyWorkbench server failed to launch.

    Examples
    --------
    Launch a server on the local computer and use the ``wb`` variable to hold the returned client.

    >>> from ansys.workbench.core import launch_workbench
    >>> wb = launch_workbench()
    """

    def __init__(
        self,
        show_gui=True,
        version=None,
        client_workdir=None,
        server_workdir=None,
        port=-1,
        use_insecure_connection=False,
        host=None,
        username=None,
        password=None,
    ):
        self._launcher = Launcher()
        port, security = self._launcher.launch(
            version,
            show_gui,
            server_workdir,
            port,
            use_insecure_connection,
            host,
            username,
            password,
        )
        if port is None or port <= 0:
            raise Exception("Failed to launch Ansys Workbench service.")
        if use_insecure_connection:
            print(
                "Using insecure connection is not recommended. "
                "Please see the documentation for your installed "
                "product for additional information."
            )
        super().__init__(port, client_workdir, host, security)
        atexit.register(self.exit)
        self._exited = False

    def exit(self):
        """Terminate the Workbench server and disconnect the client."""
        if self._exited:
            return
        self.run_script_string("Reset()")
        if self.server_version >= 252:
            self.run_script_string("internal_wbexit()")
        else:
            self._launcher.exit()
        super().exit()
        self._exited = True


def launch_workbench(
    show_gui=True,
    version=None,
    client_workdir=None,
    server_workdir=None,
    port=-1,
    use_insecure_connection=False,
    host=None,
    username=None,
    password=None,
):
    """Launch PyWorkbench server on the local machine or a remote Windows machine.

    This method launch a Workbench server on the local machine or a remote Windows machine
    and creates a PyWorkbench client that connects to the server.

    Parameters
    ----------
    show_gui : bool, default: True
        Weather to launch Workbench in UI mode.
    version : str, default: None
        Workbench version to launch. It must be a 3-digit version that is "242" or later.
        If None, the latest version available will be used.
    client_workdir : str, default: None
        Path to a writable directory on the client computer. The default is ``None``,
        in which case the system temp directory is used.
    server_workdir : str, None
        Path to a writable directory on the server computer. The default is ``None``,
        in which case the user preference for the Workbench temporary file folder is used.
    port : int, optional
        Port to use for the launched server. If not specified, port is auto determined.
    use_insecure_connection : bool, default: False
        whether to use insecure connection between the server and clients
    host : str, None
        Server computer's name or IP address. The default is ``None`` for launching on the
        local computer.
    username : str, None
        User's login name on the server computer. The default is ``None`` for launching on
        the local computer.
    password : str, None
        User's password on the server computer. The default is ``None`` for launching on
        the local computer.

    Returns
    -------
    WorkbenchClient
        Instance of the PyWorkbench client that is connected to the launched server.

    Examples
    --------
    Launch a server on the local computer and use the ``wb`` variable to hold the returned client.

    >>> from ansys.workbench.core import launch_workbench
    >>> wb = launch_workbench()

    """
    return LaunchWorkbench(
        show_gui,
        version,
        client_workdir,
        server_workdir,
        port,
        use_insecure_connection,
        host,
        username,
        password,
    )


def connect_workbench(port, client_workdir=None, host=None, security="mtls"):
    """Create a PyWorkbench client that connects to an already running Workbench server.

    Parameters
    ----------
    port : int
        Port used by the server.
    client_workdir : str, default: None
        Path to a writable directory on the client computer. The default is ``None``,
        in which case the system temp directory is used.
    host : str, default: None
        Server computer's name or IP address. The default is ``None`` for the local computer.
    security : str among 'mtls', 'wnua', 'insecure', default: 'mtls'
        Transport mode used for connection security. The default is `mtls`.

    Returns
    -------
    WorkbenchClient
        Instance of the PyWorkbench client that is connected to the server.

    Examples
    --------
    Connect to a server at port 32588 on localhost and use the ``wb`` variable to hold the
    returned client.

    >>> from ansys.workbench.core import connect_workbench
    >>> wb = connect_workbench(port = 32588)
    """
    return ClientWrapper(port, client_workdir, host, security)


__all__ = ["launch_workbench", "connect_workbench"]



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module for downloading example data from the ``example-data`` repository."""

import logging
import os
import shutil
import urllib.request


class ExampleData:
    """Provides for downloading example data from the ``example-data`` repository."""

    def _get_file_url(relative_file_path):  # noqa: N805
        """Get the URL of the file in the ``example-data`` repository.

        Parameters
        ----------
        filename : str
            Name of the file.
        dirname : str
            Name of the directory containing the file.

        Returns
        -------
        str
            URL of the file in the ``example-data`` repository.
        """
        return f"https://github.com/ansys/example-data/raw/master/pyworkbench/{relative_file_path}"

    def __retrieve_file(url, local_file_path):  # noqa: N805
        """Download the file from the URL.

        Parameters
        ----------
        url : str
            URL of the file.
        local_file_path : str
            Local path to save the file to.

        Raises
        ------
        Exception
            If the URL is not accessible.
        """
        logging.info(f"Downloading {url} from ``example-data`` repository ...")

        # Ignoring bandit error as we need to use urllib to download the file
        with urllib.request.urlopen(url) as in_stream:  # nosec: B310
            if in_stream.code != 200:
                raise Exception("error getting the url, code " + str(in_stream.code))
            with open(local_file_path, "wb") as out_file:
                shutil.copyfileobj(in_stream, out_file)
        logging.info(f"Downloaded the file as {local_file_path}.")

    def download(relative_file_path, local_dir_path):  # noqa: N805
        """Download the file from the ``example-data`` repository.

        Parameters
        ----------
        relative_file_path : str
            File path relative to the ``pyworkbench`` folder in the ``example-data`` repository.
        local_dir_path : str
           Local path to the directory to save the file to.

        Returns
        -------
        str
            Name of the downloaded file.
        """
        url = ExampleData._get_file_url(relative_file_path)
        downloaded_file_name = os.path.basename(relative_file_path)
        local_file_path = os.path.join(local_dir_path, downloaded_file_name)
        ExampleData.__retrieve_file(url, local_file_path)
        return downloaded_file_name


__all__ = []



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Workbench client module for PyWorkbench."""

import glob
import json
import logging
import logging.handlers
from logging.handlers import WatchedFileHandler
import os
import re
import time

import tqdm

from ansys.api.workbench.v0 import workbench_pb2 as wb
from ansys.api.workbench.v0.workbench_pb2_grpc import WorkbenchServiceStub
from ansys.tools.common.cyberchannel import create_channel
from ansys.workbench.core.example_data import ExampleData


class WorkbenchClient:
    """Functions of a PyWorkbench client.

    Parameters
    ----------
    local_workdir : str
        Local working directory for the client.
    server_host : str
        Hostname or IP address of the server.
    server_port : int
        Port number of the server.
    server_security : string
        Security mode of the server.
        Options are: "insecure", "uds", "wnua", "mtls"
    """

    def __init__(self, local_workdir, server_host, server_port, server_security):
        """Create a Workbench client."""
        self.workdir = local_workdir
        self._server_host = server_host
        self._server_port = server_port
        self._server_security = server_security
        self._server_version = -1
        self.__init_logging()

    @property
    def server_version(self):
        """The Workbench version of the connected server."""
        if self._server_version <= 0:
            ver_str = self.run_script_string("""import json
wb_script_result=json.dumps(GetFrameworkVersion())""")
            ver_str = ver_str.replace(".", "")
            self._server_version = int(ver_str)
        return self._server_version

    def __enter__(self):
        """Connect to the server when entering a context."""
        self._connect()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Disconnect from the server when exiting a context."""
        self._disconnect()

    def _connect(self):
        """Connect to the server."""
        self.channel = create_channel(
            host=self._server_host,
            port=self._server_port,
            transport_mode=self._server_security,
            certs_dir=None,
            grpc_options=None,
        )
        self.stub = WorkbenchServiceStub(self.channel)
        logging.info(
            f"connected to the WB server at {self._server_host}:{self._server_port} "
            "using {self._server_security} connection"
        )

    def _disconnect(self):
        """Disconnect from the server."""
        if self.channel:
            self.channel.close()
            self.channel = None
            self.stub = None
            self._server_version = -1
            logging.info("Disconnected from the Workbench server")

    def _is_connected(self):
        """Return whether this client is connected to the server."""
        return self.channel is not None

    def __init_logging(self):
        """Initialize logging for the client."""
        self._logger = logging.getLogger("WB")
        self._logger.setLevel(logging.DEBUG)
        self._logger.propagate = False
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(logging.Formatter("%(levelname)s: %(message)s"))
        stream_handler.setLevel(logging.WARNING)
        self._logger.addHandler(stream_handler)
        self.__log_console_handler = stream_handler

    def set_console_log_level(self, log_level):
        """Set the log filter level for the client console.

        Parameters
        ----------
        log_level : str, default: "error"
            Level of logging. Options are "critical", "debug", "error", "info", and "warning".
        """
        self.__log_console_handler.setLevel(WorkbenchClient.__to_python_log_level(log_level))

    def set_log_file(self, log_file):
        """Set a local log file for the Workbench server log.

        Create a local log file if one does not exist and append it to the existing log file.

        Parameters
        ----------
        log_file : str
            Path to a local file to use for logging.
        """
        self.reset_log_file()

        file_handler = WatchedFileHandler(log_file)
        file_handler.setFormatter(logging.Formatter("%(levelname)s: %(message)s"))
        file_handler.setLevel(logging.DEBUG)
        self._logger.addHandler(file_handler)
        self.__log_file_handler = file_handler

    def reset_log_file(self):
        """No longer use the current log file for the Workbench server log."""
        if self.__log_file_handler is None:
            return
        self.__log_file_handler.close()
        self._logger.removeHandler(self.__log_file_handler)
        self.__log_file_handler = None

    __log_file_handler = None
    __log_console_handler = None

    def run_script_string(self, script_string, args=None, log_level="error"):
        """Run a script as given in the input string on the server.

        Parameters
        ----------
        script_string : str
            String containing the content of the script to run.
        args : dictionary of script variable names and values
            Variables in the script specified as $$varname%%50%% will be converted to variable
            values or use the default value - 50 in the example.
        log_level : str, default: "error"
            Level of logging. Options are "critical", "debug", "error", "info", and "warning".

        Returns
        -------
        str
            Output defined in the script.

        Examples
        --------
        Run a Workbench script, given in a string, that returns the name of
        a newly created system.

        >>> wb.run_script_string(r'''import json
        wb_script_result=json.dumps(GetTemplate(TemplateName="FLUENT").CreateSystem().Name)
        ''')

        """
        if not self._is_connected():
            logging.error("Workbench client is not yet connected to a server.")
            return None
        updated_script_string = script_string
        if args and len(args) > 0:
            if any(not re.match(r"^\w+$", arg) for arg in args.keys()):
                logging.error("script argument name contains illegal character.")
                return None
            for arg_name in args:
                updated_script_string = re.sub(
                    r"\$\$" + arg_name + r"%%((?!%%).)*%%",
                    str(args[arg_name]),
                    updated_script_string,
                )
        updated_script_string = re.sub(r"\$\$\w+%%(((?!%%).)*)%%", r"\1", updated_script_string)
        request = wb.RunScriptRequest(
            content=updated_script_string,
            log_level=WorkbenchClient.__to_server_log_level(log_level),
        )
        for response in self.stub.RunScript(request):
            if response.log and response.log.messages and len(response.log.messages) > 0:
                for log_entry in response.log.messages:
                    self.__python_logging(log_entry.level, log_entry.message)
            if response.result:
                if response.result.error:
                    logging.error("Error when running the script: " + response.result.error)
                    return None
                elif response.result.result:
                    logging.info("The script has finisished.")
                    return json.loads(response.result.result)

    def run_script_file(self, script_file_name, args=None, log_level="error"):
        """Run a script file on the server.

        Parameters
        ----------
        script_file_name : str
            Name of the script file to run. The script file should be located in the client
            working directory
        args : dictionary of script variable names and values
            Variables in the script specified as $$varname%%50%% will be converted to variable
            values or use the default value - 50 in the example.
        log_level : str, default: "error"
            Level of logging. Options are "critical", "debug", "error", "info", and "warning".

        Returns
        -------
        str
            Output defined in the script.
        """
        if not self._is_connected():
            logging.error("Workbench client is not yet connected to a server")
            return None
        script_path = os.path.join(self.workdir, script_file_name)
        with open(script_path, encoding="utf-8-sig") as sf:
            script_string = sf.read()
        return self.run_script_string(script_string, args, log_level)

    def upload_file(self, *file_list, show_progress=True):
        """Upload one or more files from the client to the server.

        Parameters
        ----------
        file_list : list[str]
            List of paths to the one or more local files to upload. The
            wildcard characters "?" and "*" are supported.
        show_progress : bool, default: True
            Whether to show a progress bar during the upload.

        Returns
        -------
        list[str]
            Names of the uploaded files.
        """
        if not self._is_connected():
            logging.error("Workbench client is not yet connected to a server.")
            return
        requested = []
        for file_pattern in file_list:
            if "*" in file_pattern or "?" in file_pattern:
                if not os.path.isabs(file_pattern):
                    file_pattern = os.path.join(self.workdir, file_pattern)
                requested.extend(glob.glob(file_pattern))
            else:
                requested.append(file_pattern)
        existing_files = []
        nonexisting_files = []
        for file_name in requested:
            if not os.path.isabs(file_name):
                file_name = os.path.join(self.workdir, file_name)
            if os.path.isfile(file_name):
                existing_files.append(file_name)
            else:
                nonexisting_files.append(file_name)
        if len(nonexisting_files) > 0:
            logging.warning(
                "The following files do not exist and are skipped: " + "\n".join(nonexisting_files)
            )
        for file_path in existing_files:
            logging.info(f"uploading file {file_path}")
            response = self.stub.UploadFile(self.__upload_iterator(file_path, show_progress))
            if response.error:
                logging.error("Error during file upload: " + response.error)
            else:
                logging.info(
                    "A file is uploaded to the server with the name: " + response.file_name
                )

    def __upload_iterator(self, file_path, show_progress):
        dir_path, file_name = os.path.split(file_path)
        yield wb.UploadFileRequest(file_name=file_name)

        pbar = None
        if show_progress:
            bytes = os.path.getsize(file_path)
            pbar = tqdm.tqdm(
                total=bytes,
                desc=f"Uploading {file_name}",
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
            )

        chunk_size = 65536  ## 64 kb
        with open(file_path, mode="rb") as f:
            while True:
                chunk = f.read(chunk_size)
                if chunk:
                    if pbar is not None:
                        pbar.update(len(chunk))
                    yield wb.UploadFileRequest(file_content=chunk)
                else:
                    if pbar is not None:
                        pbar.close()
                    return

    def upload_file_from_example_repo(self, relative_file_path, show_progress=True):
        """Upload a file from the Ansys ``example-data`` repository to the server.

        Parameters
        ----------
        relative_file_path : str
            File path relative to the ``pyworkbench`` folder in the ``example-data`` repository.
        show_progress : bool, default: True
            Whether to show a progress bar during the upload.
        """
        if not self._is_connected():
            logging.error("Workbench client is not yet connected to a server.")
            return
        downloaded = ExampleData.download(relative_file_path, self.workdir)
        self.upload_file(downloaded, show_progress=show_progress)

    def download_file(self, file_name, show_progress=True, target_dir=None):
        """Download one or more files from the server.

        Parameters
        ----------
        file_name : str
            Name of the file. File must be located in the server's working directory.
            The wildcard characters "?" and "*" are supported. A ZIP file is automatically
            generated and downloaded when multiple files are specified.
        show_progress : bool, default: True
            Whether to show a progress bar during the download.
        target_dir : str, default: None
            Path to a local directory to download the files to. The default is ``None``,
            in which case the client working directory is used.


        Returns
        -------
        str
            Name of the downloaded file.
        """
        if not self._is_connected():
            logging.error("Workbench client is not yet connected to a server.")
            return None
        request = wb.DownloadFileRequest(file_name=file_name)
        file_name = file_name.replace("*", "_").replace("?", "_")
        td = target_dir
        if td is None:
            td = self.workdir
        file_path = os.path.join(td, file_name)
        pbar = None
        started = False
        for response in self.stub.DownloadFile(request):
            if response.error:
                logging.error("Error during file download: " + response.error)
                return None
            if response.file_info:
                if response.file_info.is_archive:
                    file_name += ".zip"
                    file_path += ".zip"
                if response.file_info.file_size > 0:
                    if show_progress:
                        pbar = tqdm.tqdm(
                            total=response.file_info.file_size,
                            desc=f"Downloading {file_name}",
                            unit="B",
                            unit_scale=True,
                            unit_divisor=1024,
                        )
            if response.file_content:
                size = len(response.file_content)
                if size > 0:
                    if not started:
                        if os.path.exists(file_path):
                            os.remove(file_path)
                        started = True

                    try_more = 3
                    while try_more > 0:
                        try:
                            with open(file_path, mode="ab") as f:
                                f.write(response.file_content)
                            try_more = 0
                        except PermissionError:
                            # intermittent "Permission Denied" error can happen
                            if try_more <= 0:
                                raise
                            try_more -= 1
                            time.sleep(0.1)

                    if pbar is not None:
                        pbar.update(size)
        logging.info(f"Dwnloaded the file {file_name}.")
        if pbar is not None:
            pbar.close()
        return file_name

    def download_project_archive(
        self, archive_name, include_solution_result_files=True, show_progress=True
    ):
        """Create and download the project archive.

        Parameters
        ----------
        archive_name : str
            Name of the project archive to use, without the file extension.
        include_solution_result_files : bool, default: True
            Whether to include solution and result files in the archive.
        show_progress : bool, default: True
            Whether to show a progress bar during the download.
        """
        if not re.match(r"^\w+$", archive_name):
            logging.error("archive name should contain only alphanumeric characters")
            return
        script = f"""import os
import json
successful = False
wd = GetServerWorkingDirectory()
if os.path.basename(GetProjectFile()).StartsWith("wbnew."):
    Save(FilePath=os.path.join(wd, "{archive_name}.wbpj"), Overwrite=True)
else:
    Save(Overwrite=True)
Archive(FilePath=os.path.join(wd, "{archive_name}.wbpz"),
    IncludeSkippedFiles={include_solution_result_files})
successful = True
wb_script_result =json.dumps(successful)
"""
        archive_created = self.run_script_string(script)
        if not archive_created:
            logging.error(
                (
                    "Failed to create the project archive. "
                    "Make sure that the solver PyAnsys sessions are closed."
                )
            )
            return
        self.download_file(archive_name + ".wbpz", show_progress=show_progress)

    def __python_logging(self, log_level, msg):
        """Log a message with the given log level.

        Parameters
        ----------
        log_level : int
            log level. Options are ``logging.CRITICAL``, ``logging.DEBUG``, ``logging.ERROR``,
            ``logging.INFO``, and ``logging.WARNING``.

        msg : str
            Message to log.
        """
        if log_level == wb.LOG_DEBUG:
            self._logger.debug(msg)
        elif log_level == wb.LOG_INFO:
            self._logger.info(msg)
        elif log_level == wb.LOG_WARNING:
            self._logger.warn(msg)
        elif log_level == wb.LOG_ERROR:
            self._logger.error(msg)
        elif log_level == wb.LOG_FATAL:
            self._logger.fatal(msg)

    @staticmethod
    def __to_python_log_level(log_level):
        """Convert the given log level to the corresponding Python log level.

        Parameters
        ----------
        log_level : str
            level of logging: options are "debug", "info", "warning", "error", "critical"

        Returns
        -------
        int
            the corresponding Python log level
        """
        log_level = log_level.lower()
        for level_name, server_level in WorkbenchClient.__log_levels.items():
            if log_level in level_name:
                return server_level[1]
        return logging.NOTSET

    @staticmethod
    def __to_server_log_level(log_level):
        """Convert the given log level to the corresponding server log level.

        Parameters
        ----------
        log_level : str
            Level of logging. Options are "critical", "debug", "error", "info", and "warning".

        Returns
        -------
        int
            Corresponding server log level.
        """
        log_level = log_level.lower()
        for level_name, server_level in WorkbenchClient.__log_levels.items():
            if log_level in level_name:
                return server_level[0]
        return wb.LOG_NONE

    __log_levels = {
        "none null": (wb.LOG_NONE, logging.NOTSET),
        "debug": (wb.LOG_DEBUG, logging.DEBUG),
        "information": (wb.LOG_INFO, logging.INFO),
        "warning": (wb.LOG_WARNING, logging.WARNING),
        "error": (wb.LOG_ERROR, logging.ERROR),
        "fatal critical": (wb.LOG_FATAL, logging.CRITICAL),
    }

    def start_mechanical_server(self, system_name, port=0):
        """Start the PyMechanical server for the given system in the Workbench project.

        Parameters
        ----------
        system_name : str
            Name of the system in the Workbench project.
        port : int, default: 0
            Port to use for mechanical server if possible

        Returns
        -------
        int
            Port of the PyMechanical server to use to start the PyMechanical client.

        Examples
        --------
        Start a PyMechanical session for the given system.

        >>> from ansys.mechanical.core import connect_to_mechanical
        >>> server_port=wb.start_mechanical_server(system_name=mech_system_name)
        >>> mechanical = connect_to_mechanical(port=server_port)

        """
        port_arg = ""
        if port > 0:
            port_arg = ", Port=" + str(port)
        pymech_port = self.run_script_string(
            f"""import json
if float(GetFrameworkVersion()) >= 25.2:
    server_port=LaunchMechanicalServerOnSystem(SystemName="{system_name}{port_arg}")
else:
    server_port=LaunchMechanicalServerOnSystem(SystemName="{system_name}")
wb_script_result=json.dumps(server_port)
"""
        )
        return pymech_port

    def stop_mechanical_server(self, system_name):
        """Stop the PyMechanical server for the given system in the Workbench project.

        Parameters
        ----------
        system_name : str
            Name of the system in the Workbench project.

        Examples
        --------
        Stop the PyMechanical session for the given system.

        >>> wb.stop_mechanical_server(system_name=mech_system_name)

        """
        self.run_script_string(f"""if float(GetFrameworkVersion()) >= 25.2:
    StopMechanicalServerOnSystem(SystemName="{system_name}")
""")

    def start_fluent_server(self, system_name):
        """Start the PyFluent server for the given system in the Workbench project.

        Parameters
        ----------
        system_name : str
            Name of the system in the Workbench project.

        Returns
        -------
        str
            Path to the local file containing the PyFluent server information that
            can be used to start a PyFluent client.

        Examples
        --------
        Start a PyFluent session for the given system.

        >>> import ansys.fluent.core as pyfluent
        >>> server_info_file=wb.start_fluent_server(system_name=fluent_sys_name)
        >>> fluent=pyfluent.connect_to_fluent(server_info_file_name=server_info_file)

        """
        server_info_file_name = self.run_script_string(
            f"""import json
server_info_file=LaunchFluentServerOnSystem(SystemName="{system_name}")
wb_script_result=json.dumps(server_info_file)
"""
        )
        local_copy = os.path.join(self.workdir, server_info_file_name)
        if os.path.exists(local_copy):
            os.remove(local_copy)
        self.download_file(server_info_file_name, show_progress=False)
        return local_copy

    def stop_fluent_server(self, system_name):
        """Stop the Fluent server for the given system in the Workbench project.

        Parameters
        ----------
        system_name : str
            Name of the system in the Workbench project.

        Examples
        --------
        Stop the Fluent session for the given system.

        >>> wb.stop_fluent_server(system_name=mech_system_name)

        """
        self.run_script_string(f"""if float(GetFrameworkVersion()) >= 25.2:
    StopFluentServerOnSystem(SystemName="{system_name}")
""")

    def start_sherlock_server(self, system_name):
        """Start the PySherlock server for the given system in the Workbench project.

        Parameters
        ----------
        system_name : str
            Name of the system in the Workbench project.

        Returns
        -------
        int
            Port of the PySherlock server to use to start a PySherlock client.

        Examples
        --------
        Start a PySherlock session for the given system.

        >>> from ansys.sherlock.core import pysherlock
        >>> server_port=wb.start_sherlock_server(system_name=sherlock_system_name)
        >>> sherlock = pysherlock.connect_grpc_channel(port=server_port)
        >>> sherlock.common.check()

        """
        pysherlock_port = self.run_script_string(
            f"""import json
server_port=LaunchSherlockServerOnSystem(SystemName="{system_name}")
wb_script_result=json.dumps(server_port)
"""
        )
        return pysherlock_port

    def stop_sherlock_server(self, system_name):
        """Stop the Sherlock server for the given system in the Workbench project.

        Parameters
        ----------
        system_name : str
            Name of the system in the Workbench project.

        Examples
        --------
        Stop the Sherlock session for the given system.

        >>> wb.stop_sherlock_server(system_name=mech_system_name)

        """
        self.run_script_string(f"""if float(GetFrameworkVersion()) >= 25.2:
    StopSherlockServerOnSystem(SystemName="{system_name}")
""")


__all__ = ["WorkbenchClient"]



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module for launching PyWorkbench."""

import ctypes
import logging
import os
import platform

# subprocess is used to launch Workbench on local Linux machine, excluding bandit warning
import subprocess  # nosec: B404
import time
import uuid


class Launcher:
    """Launch a Workbench server on a local or remote machine.

    Note that launching a server on a remote Linux machine, or on a remote Windows
    machine from a Linux machine, is not supported.

    Raises
    ------
    Exception
        If the wmi module on Windows, or the ctypes module on Linux, is not available
    """

    def __init__(self):
        self._wmi = None
        self._libc = None
        if platform.system() == "Windows":
            import wmi

            self._wmi = wmi
            if not self._wmi:
                raise Exception(
                    "Required Python module WMI does not exist or does not load properly."
                )
        else:
            try:
                self._libc = ctypes.CDLL("libc.so.6")
            except Exception:
                self._libc = None
            if not self._libc:
                raise Exception(
                    "Required Python module ctypes does not exist or does not load "
                    "libc.so.6 properly."
                )

        self._wmi_connection = None
        self._process_id = -1

    def launch(
        self,
        version=None,
        show_gui=True,
        server_workdir=None,
        port_to_use=-1,
        use_insecure_connection=False,
        host=None,
        username=None,
        password=None,
    ):
        """Launch PyWorkbench server on the local or a remote computer.

        Parameters
        ----------
        version : str
            Workbench version to launch. It must be a three-digit version, such as ``242`` or later.
            The latest version available will be used if None.
        show_gui : bool, default: True
            Whether to launch Workbench in UI mode.
        server_workdir : str, default: None
            Path to a writable directory on the server. The default is ``None``,
            in which case the user preference for the Workbench temporary file folder is used.
        use_insecure_connection : bool, default: False
            whether to use insecure connection between the server and clients
        host : str, default: None
            Name or IP address of the server. The default is ``None``, which launches Workbench
            on the local computer.
        username : str, default: None
            User's login name on the server. The default is ``None``, which launches Workbench
            on the local computer.
        password : str, default: None
            User's password on the server. The default is ``None``, which launches Workbench
            on the local computer.

        Raises
        ------
        Exception
            If the Ansys version string is invalid.
            If the host is given but the username or password is missing.
            If remote launching is attempted from or onto a Linux machine
            If the wmi service on the remote Windows machine fails
            If the Ansys installation is not found.
        """
        if version and (
            len(version) != 3
            or not version.isdigit()
            or version[0] not in ["2", "3"]
            or version[2] not in ["1", "2"]
            or int(version) < 242
        ):
            raise Exception("Invalid Ansys version: " + version)

        if host and not self._wmi:
            raise Exception(
                "Launching PyWorkbench on a remote machine from Linux is not supported."
            )

        if host and (not username or not password):
            raise Exception(
                "Username and password must be specified "
                "to launch PyWorkbench on a remote machine."
            )

        security = "mtls"
        if use_insecure_connection:
            security = "insecure"
        elif not host and self._wmi:
            security = "wnua"

        if self._wmi:
            try:
                if not host:
                    self._wmi_connection = self._wmi.WMI()
                else:
                    self._wmi_connection = self._wmi.WMI(host, user=username, password=password)
                logging.info("Host connection is established.")
            except self._wmi.x_wmi:
                if host:
                    raise Exception(
                        f"Failed in initializing WMI service on the remote machine {host}. "
                        "Make sure that the remote host is a Windows machine and that "
                        "the correct credential is used."
                    )
                else:
                    raise Exception("Failed in initializing WMI service on the local computer.")

        ansys_install_path = None
        if version:
            ansys_install_path = self.__getenv("AWP_ROOT" + version)
        else:
            for version_to_check in ["272", "271", "261", "252", "251", "242"]:
                ansys_install_path = self.__getenv("AWP_ROOT" + version_to_check)
                if ansys_install_path:
                    version = version_to_check
                    break
        if ansys_install_path:
            logging.info(f"Ansys installation is found at: {ansys_install_path}")
        else:
            if version:
                raise Exception(
                    (
                        f"Ansys {version} installation is not found."
                        "Make sure that environment AWP_ROOTxxx is defined."
                    )
                )
            else:
                raise Exception(
                    (
                        "No Ansys installation is not found."
                        "Make sure that environment AWP_ROOTxxx is defined."
                    )
                )

        if not host and not self._wmi and int(version) < 252:
            raise Exception("Launching PyWorkbench 25.1 or below on Linux is not supported.")

        args = []
        if platform.system() == "Windows":
            exe_loc = os.path.join("Win64", "RunWB2.exe")
        else:
            exe_loc = os.path.join("Linux64", "runwb2")
        executable = os.path.join(ansys_install_path, "Framework", "bin", exe_loc)
        args.append(executable)
        if show_gui:
            args.append("-I")
        else:
            args.append("--start-and-wait")
            args.append("-nowindow")
        args.append("-E")

        # create command string
        prefix = uuid.uuid4().hex
        cmd1 = "StartServer(EnvironmentPrefix='"
        cmd1 += prefix + "'"
        if server_workdir is not None:
            # use forward slash only to avoid escaping as command line argument
            server_workdir = server_workdir.replace("\\", "/")
            cmd1 += ",WorkingDirectory='" + server_workdir + "'"
        if port_to_use > 0:
            cmd1 += ",PortToUse=" + str(port_to_use)
        cmd2 = str(cmd1)
        cmd1 += ")"
        cmd2 += ",Security='" + security + "'"
        if host is not None:
            cmd2 += ",AllowRemoteConnection=True"
        cmd2 += ")"

        if self._wmi:
            quote_or_not = '"'  # quotes needed when constructing command line
        else:
            quote_or_not = ""
        cmd = (
            quote_or_not
            + cmd2
            + """ if __scriptingEngine__.CommandContext.AddinManager.GetAddin("""
            + """'Ansys.RemoteWB.Addin').Version.Major > 1 else """
            + cmd1
            + quote_or_not
        )
        args.append(cmd)

        command_line = " ".join(args)

        # security precaution statement
        if host is not None:
            print("""The server started will allow remote access connections to be
established, possibly permitting control of the machine and any data which resides on it.
It is highly recommended to only utilize these features on a trusted, secure network.""")

        successful = False
        process = None
        if self._wmi:
            process_startup_info = self._wmi_connection.Win32_ProcessStartup.new(ShowWindow=1)
            process_id, result = self._wmi_connection.Win32_Process.Create(
                CommandLine=command_line, ProcessStartupInformation=process_startup_info
            )
            if result == 0:
                successful = True
                self._process_id = process_id
                for p in self._wmi_connection.Win32_Process():
                    if p.ProcessId == process_id:
                        self._process = p
                        break
        else:
            # Excluding bandit warning for subprocess usage
            # as this is a controlled env
            process = subprocess.Popen(
                args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True
            )  # nosec: B603
            if process:
                successful = True
                self._process_id = process.pid
                self._process = process
        if successful:
            logging.info(f"Workbench is launched successfully with process ID {self._process_id}.")
        else:
            logging.error("Workbench failed to launch on the host.")
            return 0, security

        # retrieve server port once WB is fully up running
        port = None
        timeout = 60 * 8  # set 8 minutes as upper limit for WB startup
        start_time = time.time()
        if self._wmi:
            while True:
                port = self.__getenv("ANSYS_FRAMEWORK_SERVER_PORT")
                if port and port.startswith(prefix):
                    port = port[len(prefix) :]
                    break
                else:
                    port = None
                if time.time() - start_time > timeout:
                    logging.error("Failed to start Workbench service within reasonable timeout.")
                    break
                time.sleep(10)
        else:
            for line in process.stdout:
                line = line.rstrip()
                if line.startswith("ANSYS_FRAMEWORK_SERVER_PORT="):
                    port = line[28:]
                    if port.startswith(prefix):
                        port = port[len(prefix) :]
                        break
        if not port or int(port) <= 0:
            logging.error("Failed to retrieve the port used by Workbench service.")
            return 0, security
        logging.info("Workbench service uses port: " + port)
        return int(port), security

    def __getenv(self, key):
        value = None
        if self._wmi:
            for ev in self._wmi_connection.Win32_Environment():
                if ev.Name == key:
                    value = ev.VariableValue
        elif self._libc:
            getenv = self._libc.getenv
            getenv.restype = ctypes.c_char_p
            value = getenv(key.encode("utf-8"))
            if value:
                value = value.decode("utf-8")
        else:
            raise Exception("Code path is unexpected.")
        return value

    def exit(self):
        """End the launched Workbench server."""
        if self._process:
            try:
                if self._wmi:
                    for p in self.__collect_process_tree():
                        logging.info("Shutting down " + p.Name + " ...")
                        try:
                            p.Terminate()
                        except Exception as ex:
                            logging.info(f"Failed to terminate process {p.Name}: {ex}")
                    self._process.Terminate()
                else:
                    self._process.terminate()
            except Exception as ex:
                logging.info(f"Failed to terminate process {self._process.pid}: {ex}")

        self._wmi_connection = None
        self._process_id = -1
        self._process = None

    def __collect_process_tree(self):
        # collect parent-children mapping
        children = {self._process_id: []}
        process_by_id = {}
        for p in self._wmi_connection.Win32_Process():
            process_by_id[p.ProcessId] = p
            children.setdefault(p.ProcessId, [])
            if p.ParentProcessId is None:
                continue
            children.setdefault(p.ParentProcessId, [])
            children[p.ParentProcessId].append(p.ProcessId)

        # collect related processes
        process_ids_parent_first = []
        this_level = set([self._process_id])
        while True:
            next_level = set()
            for p in this_level:
                next_level.update(children[p])
            if len(next_level) == 0:
                break
            process_ids_parent_first.append(next_level)
            this_level = next_level

        # return the processes child-first-order
        processes_child_first = []
        for pids in reversed(process_ids_parent_first):
            for pid in pids:
                processes_child_first.append(process_by_id[pid])
        return processes_child_first
